/*
 * File name: hw09-main.c
 * Date:      2017/04/08 10:44
 * Author:    Jan Faigl
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
 
#include <termios.h> 
#include <unistd.h>  // for STDIN_FILENO
 
#include <pthread.h>

#include "prg_serial_nonblock.h"

#define MEASURE_PERIOD_MS 5000 //5 seconds

#define SERIAL_READ_TIMOUT_MS 500 //timeout for reading from serial port, quite high but due to 5 seconds alarm it is ok
 
typedef struct {
   bool quit;
   bool led;
   char send_char;
   char received_char;
   int fd; // serial port file descriptor
   int tick_counter;
   int estimated_period;
} data_t;
 
pthread_mutex_t mtx;
pthread_cond_t convar;
 
void call_termios(int reset);
 
void* input_thread(void*);
void* output_thread(void*);
void* serial_thread(void*);
void* alarm_thread(void*);
 
int main(int argc, char *argv[])
{
   data_t data = { .quit = false, .fd = -1, .send_char = ' ', .received_char = '?', .led = '?', .tick_counter = 0, .estimated_period = -1};

   const char *serial = argc > 1 ? argv[1] : "/dev/ttyACM0"; // or /dev/cuaU0;
   data.fd = serial_open(serial);

   if (data.fd == -1) {
      fprintf(stderr, "Cannot open serial port %s\n", serial);
      exit(100);
   }

   enum { INPUT, OUTPUT, SERIAL, ALARM, NUM_THREADS };
   const char *threads_names[] = { "Input", "Output", "Serial", "Alarm" };

   void* (*thr_functions[])(void*) = { input_thread, output_thread, serial_thread, alarm_thread };

   pthread_t threads[NUM_THREADS];
   pthread_mutex_init(&mtx, NULL); // initialize mutex with default attributes
   pthread_cond_init(&convar, NULL); // initialize mutex with default attributes

   call_termios(0);

   for (int i = 0; i < NUM_THREADS; ++i) {
      int r = pthread_create(&threads[i], NULL, thr_functions[i], &data);
      printf("Create thread '%s' %s\n", threads_names[i], ( r == 0 ? "OK" : "FAIL") );
   }

   int *ex;
   for (int i = 0; i < NUM_THREADS; ++i) {
      printf("Call join to the thread %s\n", threads_names[i]);
      int r = pthread_join(threads[i], (void*)&ex);
      printf("Joining the thread %s has been %s - exit value %i\n", threads_names[i], (r == 0 ? "OK" : "FAIL"), *ex);
   }

   serial_close(data.fd);
   call_termios(1); // restore terminal settings
   return EXIT_SUCCESS;
}

// - function -----------------------------------------------------------------
void call_termios(int reset)
{
   static struct termios tio, tioOld;
   tcgetattr(STDIN_FILENO, &tio);
   if (reset) {
      tcsetattr(STDIN_FILENO, TCSANOW, &tioOld);
   } else {
      tioOld = tio; //backup 
      cfmakeraw(&tio);
      tio.c_oflag |= OPOST;
      tcsetattr(STDIN_FILENO, TCSANOW, &tio);
   }
}

// - function -----------------------------------------------------------------
void* input_thread(void* d)
{
   data_t *data = (data_t*)d;
   static int r = 0;
   int c;
   bool end = false;
   bool notify_vis = false;
   while (( c = getchar()) != 'q' && !end) {
      pthread_mutex_lock(&mtx);
      switch(c) {
	 case 's':
	 case 'e':
	 case '1':
	 case '2':
	 case '3':
	 case '4':
	 case '5':
	 case 'b':
	 case 'h':
	    if (serial_putc(data->fd, c) != 1) {
	       fprintf(stderr, "ERROR: Cannot send command to Nucleo, exit program\n");
	       end = true;
	       r = 1;
	    } else {
	       data->send_char = c;
	       data->received_char = '?';
	       notify_vis = true;
	    }
	    break;
	 case 'q':
	    end = true;
	    break;
	 default: // discard all other keys
	    break;
      }
      pthread_mutex_unlock(&mtx);
   }
   fprintf(stderr, "set quit\n");
   pthread_mutex_lock(&mtx);
   data->quit = true;
   pthread_cond_broadcast(&convar);
   pthread_mutex_unlock(&mtx);
   fprintf(stderr, "Exit input thread %p\n", pthread_self());
   return &r;
}

// - function -----------------------------------------------------------------
void* output_thread(void* d)
{
   data_t *data = (data_t*)d;
   static int r = 0;
   bool q = false;
   while (!q) {
      pthread_mutex_lock(&mtx);
      q = data->quit;
      if (data->estimated_period >= 0) {
	 printf("\rLED %3s send: '%c' received: '%c', T = %4d ms, ticker = %4d", data->led ? "on" : "off", data->send_char, data->received_char, data->estimated_period, data->tick_counter);
      } else {
	 printf("\rLED %3s send: '%c' received: '%c', T = %4s ms, ticker = %4d", data->led ? "on" : "off", data->send_char, data->received_char, "???", data->tick_counter);
      }
      // printf("Alarm time: %10i   Alarm counter: %10i", data->alarm_period, data->alam_counter);
      fflush(stdout);
      pthread_cond_wait(&convar, &mtx); // wait for next event
      pthread_mutex_unlock(&mtx);
   }
   fprintf(stderr, "Exit output thread %p\n", pthread_self());
   return &r;
}

// - function -----------------------------------------------------------------
void* serial_thread(void* d) 
{
   data_t *data = (data_t*)d;
   static int r = 0;
   pthread_mutex_lock(&mtx);
   bool end = data->quit;
   pthread_mutex_unlock(&mtx);

   while (!end) {
      char c;
      int r = serial_getc_timeout(data->fd, SERIAL_READ_TIMOUT_MS, &c);
      pthread_mutex_lock(&mtx);
      bool notify_vis = true;
      if (r > 0) { // character has been read
	 switch (c) {
	    case 'x':
	       data->tick_counter += 1;
	       data->led = true;
	       data->received_char = c;
	       break;
	    case 'o':
	       data->led = false;
	       data->received_char = c;
	       break;
	    case 'a':
	       if (data->send_char == 's') {
		  data->led = true;
	       } else if (data->send_char == 'e') {
		  data->led = false;
	       }
	    case 'h':
	    case 'f':
	    case 'i':
	    case 'b':
	       data->received_char = c;
	       break;
	    default:
	       fprintf(stderr, "Warn: discar some characters\n");
	       notify_vis = false;
	       break;
	 } // end switch
      } else if (r == 0) { // read but nothing happend
      } else {  //error
	 fprintf(stderr, "Error: cannot receive data from serial port\n");
	 end = true;
      }
      if (!end) {
	 end = data->quit; // check request for end
      }
      if (notify_vis) {
	 pthread_cond_broadcast(&convar); // signal threads 
      }
      pthread_mutex_unlock(&mtx);
   } //end while (!end)
   pthread_mutex_lock(&mtx);
   data->quit = true;
   pthread_cond_broadcast(&convar); // signal threads 
   pthread_mutex_unlock(&mtx);
   fprintf(stderr, "Exit serial thread %p\n", pthread_self());
   return &r;
}

// - function -----------------------------------------------------------------
void* alarm_thread(void* d) 
{
   data_t *data = (data_t*)d;
   static int r = 0;
   pthread_mutex_lock(&mtx);
   bool q = data->quit;
   pthread_mutex_unlock(&mtx);
   while (!q) {
      usleep(MEASURE_PERIOD_MS * 1000);
      pthread_mutex_lock(&mtx);
      q = data->quit;
      if (data->tick_counter > 0) {
	 data->estimated_period = MEASURE_PERIOD_MS / (2 *data->tick_counter);
      } else {
	 data->estimated_period = -1;
      }
      data->tick_counter = 0; // reset tick counter
      pthread_cond_broadcast(&convar);
      pthread_mutex_unlock(&mtx);
   }
   fprintf(stderr, "Exit alarm thread %p\n", pthread_self());
   return &r;
}

/* end of hw09-main.c */
